#!/usr/bin/env python

import os
import sys
import tempfile
from pathlib import Path

from Bio import SearchIO, SeqIO
from tqdm import tqdm

from .args_processing import process_args
from .exceptions import OrthoFisherError
from .parser import create_parser
from .helper import (
    create_directories, read_input_files, 
    conduct_hmm_search, set_names,
    handle_single_copy_writing, handle_multi_copy_writing,
    handle_absent_writing, handle_percent_present_summary,
    check_hmmsearch_output
    )

def execute(
    fasta_file_list: str,
    hmms_file_list: str,
    evalue: float,
    percent_bitscore: float,
    output_dir: str,
    cpu: int,
    force: bool = False,
    write_all_sequences: bool = False,
    keep_hmmsearch_output: bool = False
):

    # read input files
    fasta_file_list, hmms_file_list = read_input_files(fasta_file_list, hmms_file_list)

    # create directories that will be
    # populated with output files
    create_directories(output_dir, write_all_sequences, keep_hmmsearch_output, force)

    ## loop through fasta files
    for fasta in tqdm(fasta_file_list):
        # ortholog presence absence stats will hold
        # idx0: # single-copy orthologs
        # idx1: # multi-copy orthologs
        # idx2: # absent orthologs
        ortholog_presence_absence_stats = [0, 0, 0]
        record_dict = SeqIO.index(fasta[0], "fasta")
        fasta_name = Path(fasta[0]).name
        name = f"{fasta[1]} " if len(fasta) > 1 else ""

        try:
            ## loop through HMMs
            for hmm in hmms_file_list:
                hmm_name, \
                singlecopy_name, \
                all_name, \
                short_summary_name, \
                long_summary_name, \
                hmmsearch_out = set_names(hmm[0], fasta_name, output_dir)
                created_temp_hmmsearch = False
                if not keep_hmmsearch_output:
                    fd, temp_hmmsearch_out = tempfile.mkstemp(
                        prefix=f"{fasta_name}_{hmm_name}_",
                        suffix=".tbl",
                        dir=output_dir
                    )
                    os.close(fd)
                    hmmsearch_out = temp_hmmsearch_out
                    created_temp_hmmsearch = True
                try:
                    # execute hmmsearch
                    conduct_hmm_search(hmmsearch_out, hmm[0], fasta[0], evalue, cpu)

                    # identify single copy orthologs and write them to separate scog files
                    check_hmmsearch_output(hmmsearch_out)

                    with open(hmmsearch_out, 'r') as hmmsearch_handle:
                        have_written = 0
                        for qresult in SearchIO.parse(hmmsearch_handle, 'hmmer3-tab'):
                            matches = qresult.hits
                            top_score = matches[0].bitscore

                            # remove genes with bitscores less than specified value (-b arg)
                            # of the bitscore value as the top hit in the hmmsearch
                            hits = [hit for hit in matches if hit.bitscore >= (percent_bitscore * top_score)]
                            num_hits = len(hits)

                            # if single copy
                            if num_hits == 1:
                                # write to single copy orthologous gene file
                                # write to all gene file
                                # write to long log file
                                handle_single_copy_writing(
                                    singlecopy_name,
                                    name,
                                    all_name,
                                    long_summary_name,
                                    fasta_name,
                                    hmm_name,
                                    hits,
                                    record_dict,
                                    write_all_sequences=write_all_sequences
                                )

                                # add to counter
                                ortholog_presence_absence_stats[0]+=1
                                have_written+=1

                            # if multi-copy
                            elif num_hits > 1:
                                handle_multi_copy_writing(
                                    name,
                                    all_name,
                                    long_summary_name,
                                    fasta_name,
                                    hmm_name,
                                    hits,
                                    record_dict,
                                    num_hits,
                                    write_all_sequences=write_all_sequences
                                )
                                
                                # add to counter
                                ortholog_presence_absence_stats[1]+=1
                                have_written+=1
                            
                        # if absent
                        if have_written == 0:
                            handle_absent_writing(
                                long_summary_name,
                                fasta_name,
                                hmm_name
                            )

                            # add to counter
                            ortholog_presence_absence_stats[2]+=1
                finally:
                    if created_temp_hmmsearch and os.path.exists(hmmsearch_out):
                        os.remove(hmmsearch_out)
        finally:
            record_dict.close()
        
        handle_percent_present_summary(
            ortholog_presence_absence_stats,
            fasta_name,
            short_summary_name
        )

def main(argv=None):
    """
    Function that parses and collects arguments              
    """
    # parse and assign arguments
    parser = create_parser()
    args = parser.parse_args()

    # pass to master execute function
    try:
        execute(**process_args(args))
    except OrthoFisherError as exc:
        print(str(exc))
        sys.exit(1)

if __name__ == "__main__":
    main(sys.argv[1:])



import os
import os.path
import shutil
import subprocess
from pathlib import Path

from .exceptions import HMMSearchError, OutputDirectoryExistsError

def read_input_files(
    fasta_file_list: str,
    hmms_file_list: str
):
    """
    read input files from user args
    """

    ## read input files
    with open(fasta_file_list, "r") as fasta_handle:
        fasta_file_list = [i.strip('\n').split('\t') for i in fasta_handle]
    with open(hmms_file_list, "r") as hmms_handle:
        hmms_file_list = [i.strip('\n').split('\t') for i in hmms_handle]

    return fasta_file_list, hmms_file_list

def create_directories(
    output_dir,
    write_all_sequences=True,
    keep_hmmsearch_output=True,
    force=False
):
    """
    create directories for outputting files
    """

    ## create output directories
    try:
        os.mkdir(output_dir)
    except FileExistsError:
        if not force:
            raise OutputDirectoryExistsError(
                f"{output_dir} already exists. Use --force to overwrite."
            )
        shutil.rmtree(output_dir)
        os.mkdir(output_dir)
    os.mkdir(f'{output_dir}/scog')
    if keep_hmmsearch_output:
        os.mkdir(f'{output_dir}/hmmsearch_output')
    if write_all_sequences:
        os.mkdir(f'{output_dir}/all_sequences')

def conduct_hmm_search(
    hmmsearch_out: str,
    hmm: str,
    fasta: str,
    evalue: float,
    cpu: int
):
    """
    execute hmm search using subprocess
    """
    # execute hmmsearch
    try:
        subprocess.run(
            [
                'hmmsearch', '--noali',
                '--notextw', '-E', str(evalue),
                '--cpu', str(cpu),
                '--tblout', hmmsearch_out,
                hmm, fasta
            ],
            stdout=subprocess.DEVNULL,
            stderr=subprocess.PIPE,
            text=True,
            check=True,
        )
    except subprocess.CalledProcessError as exc:
        err = exc.stderr.strip() if exc.stderr else "hmmsearch failed without stderr output."
        raise HMMSearchError(f"HMM search failed: {err}")

def check_hmmsearch_output(
    hmmsearch_out: str
):
    
    if not os.path.isfile(hmmsearch_out):
        raise HMMSearchError("HMM search failed. Check e-value is an appropriate number.")

def set_names(
    hmm: str,
    fasta_name: str,
    output_dir: str
):
    """
    set names for output files
    """
    hmm_name = Path(hmm).name
    singlecopy_name = f'{output_dir}/scog/{hmm_name}.orthofisher'
    all_name = f'{output_dir}/all_sequences/{hmm_name}.orthofisher'
    long_summary_name = f'{output_dir}/long_summary.txt'
    short_summary_name = f'{output_dir}/short_summary.txt'
    hmmsearch_out = f'{output_dir}/hmmsearch_output/{fasta_name}_{hmm_name}'

    return hmm_name, singlecopy_name, all_name, short_summary_name, long_summary_name, hmmsearch_out

def handle_single_copy_writing(
    singlecopy_name: str,
    fasta: str,
    all_name: str,
    long_summary_name: str,
    fasta_name: str,
    hmm_name: str,
    hits,
    record_dict,
    write_all_sequences=True
):
    """
    writes single copy entry to two output fasta files
    and to a long-form log file
    """
    entry_name = ''.join([fasta + hits[0]._id + '|' + str(hits[0].evalue) + '|' + str(hits[0].bitscore)])
    record_entry = record_dict[hits[0].id]
    sequence = _format_sequence_lines(record_entry.seq)

    # write single copy orthologous gene file
    with open(singlecopy_name, 'a') as f:
        f.write(">" + entry_name + '\n' + sequence + '\n')

    # write to all gene file
    if write_all_sequences:
        with open(all_name, 'a') as f:
            f.write(">" + entry_name + '\n' + sequence + '\n')

    # write to long log file
    with open(long_summary_name, 'a') as f:
        f.write(
            '\t'.join(
                [
                    fasta_name,
                    hmm_name,
                    'single-copy',
                    str(1),
                    hits[0]._id,
                    str(hits[0].evalue),
                    str(hits[0].bitscore),
                ]
            ) + '\n'
        )

def handle_multi_copy_writing(
    fasta: str,
    all_name: str,
    long_summary_name: str,
    fasta_name: str,
    hmm_name: str,
    hits,
    record_dict,
    num_hits: int,
    write_all_sequences=True
):
    """
    writes single copy entry to two output fasta files
    and to a long-form log file
    """
    with open(long_summary_name, 'a') as long_out:
        if write_all_sequences:
            with open(all_name, 'a') as all_out:
                for idx in range(num_hits):
                    entry_name = ''.join([fasta + hits[idx]._id + '|' + str(hits[idx].evalue) + '|' + str(hits[idx].bitscore)])
                    record_entry = record_dict[hits[idx].id]
                    sequence = _format_sequence_lines(record_entry.seq)

                    # write to all gene file
                    all_out.write(">" + entry_name + '\n' + sequence + '\n')

                    # write to long log file
                    long_out.write(
                        '\t'.join(
                            [
                                fasta_name,
                                hmm_name,
                                'multi-copy',
                                str(num_hits),
                                hits[idx]._id,
                                str(hits[idx].evalue),
                                str(hits[idx].bitscore),
                            ]
                        ) + '\n'
                    )
        else:
            for idx in range(num_hits):
                long_out.write(
                    '\t'.join(
                        [
                            fasta_name,
                            hmm_name,
                            'multi-copy',
                            str(num_hits),
                            hits[idx]._id,
                            str(hits[idx].evalue),
                            str(hits[idx].bitscore),
                        ]
                    ) + '\n'
                )

def handle_absent_writing(
    long_summary_name: str,
    fasta_name: str,
    hmm_name: str
):
    """
    logs if hmm is absent in the genome
    """
    # write to long log file
    with open(long_summary_name, 'a') as f:
        f.write('\t'.join([fasta_name, hmm_name, 'absent', str(0), 'NA', 'NA', 'NA']) + '\n')

def handle_percent_present_summary(
    ortholog_presence_absence_stats: list,
    fasta_name: str,
    short_summary_name: str
):
    """
    write log of percent present single-copy, multi-copy, or absent
    """

    header = [
        'file_name', 'single-copy', 'multi-copy', 'absent',
        'per_single-copy', 'per_multi-copy', 'per_absent'
    ]
    total = sum(ortholog_presence_absence_stats)
    ortholog_presence_absence_per = [round(val/total, 2) for val in ortholog_presence_absence_stats]

    ortholog_presence_absence_stats = [str(i) for i in ortholog_presence_absence_stats]
    ortholog_presence_absence_per = [str(i) for i in ortholog_presence_absence_per] 

    if os.path.isfile(short_summary_name):
        with open(short_summary_name, 'a') as f:
            f.write(
                '\t'.join(
                    [fasta_name,
                    '\t'.join(ortholog_presence_absence_stats),
                    '\t'.join(ortholog_presence_absence_per)
                    ]
                ) + '\n'
            )
    else:
        with open(short_summary_name, 'a') as f:
            f.write('\t'.join(header) + '\n')

            f.write(
                '\t'.join(
                    [fasta_name,
                    '\t'.join(ortholog_presence_absence_stats),
                    '\t'.join(ortholog_presence_absence_per)
                    ]
                ) + '\n'
            )

def _format_sequence_lines(seq, width=60):
    seq_str = str(seq)
    return '\n'.join(seq_str[i:i+width] for i in range(0, len(seq_str), width))



class OrthoFisherError(Exception):
    """Base exception for orthofisher runtime errors."""


class InputValidationError(OrthoFisherError):
    """Raised when CLI/user inputs are invalid."""


class OutputDirectoryExistsError(OrthoFisherError):
    """Raised when output directory exists and overwrite is not allowed."""


class HMMSearchError(OrthoFisherError):
    """Raised when hmmsearch execution or output validation fails."""



import os.path

from .exceptions import InputValidationError

def process_args(args) -> dict:
    """
    Process args from argparser and set defaults
    """
    fasta_file_list = args.fasta
    hmms_file_list = args.hmm

    if not os.path.isfile(fasta_file_list):
        raise InputValidationError("Fasta file list does not exist")

    if not os.path.isfile(hmms_file_list):
        raise InputValidationError("HMM file does not exist")

    
    # assign optional arguments
    evalue = args.evalue if args.evalue is not None else 0.001

    percent_bitscore = args.bitscore if args.bitscore is not None else 0.85
    
    output_dir = args.output_dir or "orthofisher_output"

    cpu = args.cpu if args.cpu is not None else 2
    force = args.force or False
    verbose_output = args.verbose_output or False

    if evalue <= 0:
        raise InputValidationError("e-value must be greater than 0")

    if percent_bitscore < 0 or percent_bitscore > 1:
        raise InputValidationError("bitscore must range from 0 to 1")

    if cpu < 1:
        raise InputValidationError("cpu must be an integer greater than 0")

    write_all_sequences = verbose_output
    keep_hmmsearch_output = verbose_output

    return dict(
        fasta_file_list=fasta_file_list,
        hmms_file_list=hmms_file_list,
        evalue=evalue,
        percent_bitscore=percent_bitscore,
        output_dir=output_dir,
        cpu=cpu,
        force=force,
        write_all_sequences=write_all_sequences,
        keep_hmmsearch_output=keep_hmmsearch_output
    )
